#!/usr/bin/env python
################################################################################
#   (c) 2007 Arthur Clune
#   Author: Arthur Clune, arthur@honeynet.org.uk
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
################################################################################

     
# to run over a set of HS directories:
#
# find output |grep '/http/'|xargs grep -l '<script>' | xargs python entropy.py 

import math 
import sys  
import codecs

from BeautifulSoup import BeautifulSoup, NavigableString

# JS keywords + 'eval', 'document' and 'write'
keywords = [
    "abstract",
    "boolean",
    "break",
    "byte",
    "case",
    "catch",
    "char",
    "class",
    "const",
    "continue",
    "debugger",
    "default",
    "delete",
    "do",    
    "document",
    "double",
    "else",
    "enum",   
    "exec",
    "export",
    "extends",
    "false",
    "final",
    "finally",
    "float",
    "for",
    "function",
    "goto",
    "if",
    "implements",
    "import",
    "in",
    "instanceof",
    "int",
    "interface",
    "long",
    "native",
    "new",
    "null",
    "package",
    "private",
    "protected",
    "public",
    "return",
    "short",
    "static",
    "super",
    "switch",
    "synchronized",
    "this",
    "throw",
    "throws",
    "transient",
    "true",
    "try",
    "typeof",
    "var",
    "void",
    "volatile",
    "while",
    "with", 
    "write",
]

def H(data):    
    """classical entropy measure"""   
    if not data:
        return 0
    entropy = 0
    for x in range(128):    # assume ASCII javascript
        p_x = float(data.count(chr(x)))/len(data)
        if p_x > 0:
            entropy += - p_x*math.log(p_x, 2)
    return entropy    
                    
def I(data):
    """ratio of JS keywords to text"""
    if not data:
        return 0   
    measure = 0
    for k in keywords:
        c = float(data.count(k))/len(data)
        measure += c   
    if measure == 0:
        return 'Undef' 
    return int(1/measure)
    
def unpack(script):  
    if not script.next:
        return
    if type(script.next) != type(NavigableString()):
        unpack(script.next)
    else: 
        s = script.next
        if H(s) > 0:
            print "\tH %s, I %s" % (H(s), I(s))                

def html_page(data):
    """find scripts in html page"""
    try:
        soup = BeautifulSoup(data)     
    except UnicodeDecodeError:
        print "Can't run BeautifulSoup over file. Skipping. Try cleaning file by hand first"
        return
    scripts = soup.findAll('script')
    if len(scripts) == 0:
        print "\tNo scripts tags found in input"
        return
    for script in scripts: 
        unpack(script)    

def process_file(file):
    try:
        f = open(file)
        data = ''.join(f.readlines())
        f.close()
    except:
        print "Can't open %s for reading" % file
        sys.exit(1)                                   
    # remove unicode for now. This is a real bodge.
    try:  
        data = codecs.ascii_encode(data, "replace")[0]
    except UnicodeDecodeError:                        
        # guff in file that isn't even unicode
        o = ''
        for c in data:
            if ord(c)<128:
                o += c  
        data = o                    
    html_page(data)
  
if __name__ == '__main__':

    if len(sys.argv)==1: 
        html = sys.stdin.read()    
    elif len(sys.argv)>=2:   
        for file in sys.argv[1:]: 
            print file
            process_file(file) 
    else:
        print "Usage: %s <file>" % sys.argv[0]
        sys.exit(1)            

        